@@ -15,6 +15,7 @@ module Data.PQueue.Internals (
   minView,
   singleton,
   insert,
+  insertBehind,
   union,
   mapMaybe,
   mapEither,
@@ -30,11 +31,11 @@ module Data.PQueue.Internals (
   seqSpine
   ) where
 
-import Control.DeepSeq
+import Control.DeepSeq (NFData(rnf), deepseq)
 
-import Data.Functor
+import Data.Functor ((<$>))
 import Data.Foldable (Foldable (foldr, foldl))
-import Data.Monoid (Monoid (..))
+import Data.Monoid (mappend)
 import qualified Data.PQueue.Prio.Internals as Prio
 
 #ifdef __GLASGOW_HASKELL__
@@ -57,14 +58,14 @@ instance (Ord a, Data a) => Data (MinQueue a) where
   gfoldl f z q  = case minView q of
     Nothing      -> z Empty
     Just (x, q') -> z insertMinQ `f` x `f` q'
-  
+
   gunfold k z c = case constrIndex c of
     1  -> z Empty
     2  -> k (k (z insertMinQ))
     _  -> error "gunfold"
-  
+
   dataCast1 x = gcast1 x
-  
+
   toConstr q
     | null q  = emptyConstr
     | otherwise  = consConstr
@@ -84,30 +85,33 @@ type BinomHeap = BinomForest Zero
 
 instance Ord a => Eq (MinQueue a) where
   Empty == Empty = True
-  MinQueue n1 x1 q1 == MinQueue n2 x2 q2 = n1 == n2 && x1 == x2 && eq' q1 q2 where
-    eq' q1 q2 = case (extractHeap q1, extractHeap q2) of
-      (Just (x1, q1'), Just (x2, q2'))
-        -> x1 == x2 && eq' q1' q2'
-      (Nothing, Nothing)
-        -> True
-      _ -> False
+  MinQueue n1 x1 q1 == MinQueue n2 x2 q2 =
+    n1 == n2 && eqExtract (x1,q1) (x2,q2)
   _ == _ = False
 
+eqExtract :: Ord a => (a, BinomHeap a) -> (a, BinomHeap a) -> Bool
+eqExtract (x1,q1) (x2,q2) =
+  x1 == x2 &&
+  case (extractHeap q1, extractHeap q2) of
+    (Just h1, Just h2) -> eqExtract h1 h2
+    (Nothing, Nothing) -> True
+    _ -> False
+
 instance Ord a => Ord (MinQueue a) where
   Empty `compare` Empty = EQ
   Empty `compare` _ = LT
   _ `compare` Empty = GT
-  MinQueue _n1 x1 q1 `compare` MinQueue _n2 x2 q2 = compare x1 x2 `mappend` cmp' q1 q2 where
-    cmp' q1 q2 = case (extractHeap q1, extractHeap q2) of
-      (Just (x1, q1'), Just (x2, q2'))
-        -> compare x1 x2 `mappend` cmp' q1' q2'
-      (Nothing, Nothing)
-        -> EQ
-      (Just{}, Nothing)
-        -> GT
-      (Nothing, Just{})
-        -> LT
-      
+  MinQueue _n1 x1 q1 `compare` MinQueue _n2 x2 q2 = cmpExtract (x1,q1) (x2,q2)
+
+cmpExtract :: Ord a => (a, BinomHeap a) -> (a, BinomHeap a) -> Ordering
+cmpExtract (x1,q1) (x2,q2) =
+  compare x1 x2 `mappend`
+  case (extractHeap q1, extractHeap q2) of
+    (Just h1, Just h2) -> cmpExtract h1 h2
+    (Nothing, Nothing) -> EQ
+    (Just _, Nothing) -> GT
+    (Nothing, Just _) -> LT
+
     -- We compare their first elements, then their other elements up to the smaller queue's length,
     -- and then the longer queue wins.
     -- This is equivalent to @comparing toAscList@, except it fuses much more nicely.
@@ -115,31 +119,31 @@ instance Ord a => Ord (MinQueue a) where
 -- We implement tree ranks in the type system with a nicely elegant approach, as follows.
 -- The goal is to have the type system automatically guarantee that our binomial forest
 -- has the correct binomial structure.
--- 
+--
 -- In the traditional set-theoretic construction of the natural numbers, we define
 -- each number to be the set of numbers less than it, and Zero to be the empty set,
 -- as follows:
--- 
+--
 -- 0 = {}  1 = {0}    2 = {0, 1}  3={0, 1, 2} ...
--- 
+--
 -- Binomial trees have a similar structure: a tree of rank @k@ has one child of each
 -- rank less than @k@.  Let's define the type @rk@ corresponding to rank @k@ to refer
 -- to a collection of binomial trees of ranks @0..k-1@.  Then we can say that
--- 
+--
 -- > data Succ rk a = Succ (BinomTree rk a) (rk a)
--- 
+--
 -- and this behaves exactly as the successor operator for ranks should behave.  Furthermore,
 -- we immediately obtain that
--- 
+--
 -- > data BinomTree rk a = BinomTree a (rk a)
--- 
+--
 -- which is nice and compact.  With this construction, things work out extremely nicely:
--- 
+--
 -- > BinomTree (Succ (Succ (Succ Zero)))
--- 
+--
 -- is a type constructor that takes an element type and returns the type of binomial trees
 -- of rank @3@.
-data BinomForest rk a = Nil | Skip (BinomForest (Succ rk) a) | 
+data BinomForest rk a = Nil | Skip (BinomForest (Succ rk) a) |
   Cons {-# UNPACK #-} !(BinomTree rk a) (BinomForest (Succ rk) a)
 
 data BinomTree rk a = BinomTree a (rk a)
@@ -174,7 +178,7 @@ getMin :: MinQueue a -> Maybe a
 getMin (MinQueue _ x _) = Just x
 getMin _                = Nothing
 
--- | Retrieves the minimum element of the queue, and the queue stripped of that element, 
+-- | Retrieves the minimum element of the queue, and the queue stripped of that element,
 -- or 'Nothing' if passed an empty queue.
 minView :: Ord a => MinQueue a -> Maybe (a, MinQueue a)
 minView Empty = Nothing
@@ -186,10 +190,15 @@ minView (MinQueue n x ts) = Just (x, case extractHeap ts of
 singleton :: a -> MinQueue a
 singleton x = MinQueue 1 x Nil
 
--- | Amortized /O(1)/, worst-case /O(log n)/.  Insert an element into the priority queue.  
+-- | Amortized /O(1)/, worst-case /O(log n)/.  Insert an element into the priority queue.
 insert :: Ord a => a -> MinQueue a -> MinQueue a
 insert = insert' (<=)
 
+-- | Amortized /O(1)/, worst-case /O(log n)/.  Insert an element into the priority queue,
+--   putting it behind elements that compare equal to the inserted one.
+insertBehind :: Ord a => a -> MinQueue a -> MinQueue a
+insertBehind = insert' (<)
+
 -- | Amortized /O(log (min(n,m)))/, worst-case /O(log (max (n,m)))/.  Take the union of two priority queues.
 union :: Ord a => MinQueue a -> MinQueue a -> MinQueue a
 union = union' (<=)
@@ -235,24 +244,24 @@ foldlAsc f z (MinQueue _ x ts) = foldlUnfold f (z `f` x) extractHeap ts
 {-# INLINE foldlUnfold #-}
 -- | @foldlUnfold f z suc s0@ is equivalent to @foldl f z (unfoldr suc s0)@.
 foldlUnfold :: (c -> a -> c) -> c -> (b -> Maybe (a, b)) -> b -> c
-foldlUnfold f z suc s0 = unf z s0 where
+foldlUnfold f z0 suc s0 = unf z0 s0 where
   unf z s = case suc s of
     Nothing      -> z
     Just (x, s') -> unf (z `f` x) s'
 
 insert' :: LEq a -> a -> MinQueue a -> MinQueue a
 insert' _ x Empty = singleton x
-insert' (<=) x (MinQueue n x' ts)
-  | x <= x'   = MinQueue (n+1) x (incr (<=) (tip x') ts)
-  | otherwise = MinQueue (n+1) x' (incr (<=) (tip x) ts)
+insert' le x (MinQueue n x' ts)
+  | x `le` x' = MinQueue (n+1) x (incr le (tip x') ts)
+  | otherwise = MinQueue (n+1) x' (incr le (tip x) ts)
 
 {-# INLINE union' #-}
 union' :: LEq a -> MinQueue a -> MinQueue a -> MinQueue a
 union' _ Empty q = q
 union' _ q Empty = q
-union' (<=) (MinQueue n1 x1 f1) (MinQueue n2 x2 f2)
-  | x1 <= x2  = MinQueue (n1 + n2) x1 (carry (<=) (tip x2) f1 f2)
-  | otherwise = MinQueue (n1 + n2) x2 (carry (<=) (tip x1) f1 f2)
+union' le (MinQueue n1 x1 f1) (MinQueue n2 x2 f2)
+  | x1 `le` x2 = MinQueue (n1 + n2) x1 (carry le (tip x2) f1 f2)
+  | otherwise  = MinQueue (n1 + n2) x2 (carry le (tip x1) f1 f2)
 
 -- | Takes a size and a binomial forest and produces a priority queue with a distinguished global root.
 extractHeap :: Ord a => BinomHeap a -> Maybe (a, BinomHeap a)
@@ -262,25 +271,25 @@ extractHeap ts = case extractBin (<=) ts of
 
 -- | A specialized type intended to organize the return of extract-min queries
 -- from a binomial forest.  We walk all the way through the forest, and then
--- walk backwards.  @Extract rk a@ is the result type of an extract-min 
+-- walk backwards.  @Extract rk a@ is the result type of an extract-min
 -- operation that has walked as far backwards of rank @rk@ -- that is, it
 -- has visited every root of rank @>= rk@.
--- 
+--
 -- The interpretation of @Extract minKey children forest@ is
--- 
+--
 --   * @minKey@ is the key of the minimum root visited so far.  It may have
---     any rank @>= rk@.  We will denote the root corresponding to 
+--     any rank @>= rk@.  We will denote the root corresponding to
 --     @minKey@ as @minRoot@.
---   
---   * @children@ is those children of @minRoot@ which have not yet been 
---     merged with the rest of the forest. Specifically, these are 
+--
+--   * @children@ is those children of @minRoot@ which have not yet been
+--     merged with the rest of the forest. Specifically, these are
 --     the children with rank @< rk@.
---   
---   * @forest@ is an accumulating parameter that maintains the partial 
---     reconstruction of the binomial forest without @minRoot@. It is 
---     the union of all old roots with rank @>= rk@ (except @minRoot@), 
---     with the set of all children of @minRoot@ with rank @>= rk@.  
---     Note that @forest@ is lazy, so if we discover a smaller key 
+--
+--   * @forest@ is an accumulating parameter that maintains the partial
+--     reconstruction of the binomial forest without @minRoot@. It is
+--     the union of all old roots with rank @>= rk@ (except @minRoot@),
+--     with the set of all children of @minRoot@ with rank @>= rk@.
+--     Note that @forest@ is lazy, so if we discover a smaller key
 --     than @minKey@ later, we haven't wasted significant work.
 data Extract rk a = Extract a (rk a) (BinomForest rk a)
 data MExtract rk a = No | Yes {-# UNPACK #-} !(Extract rk a)
@@ -290,47 +299,47 @@ incrExtract (Extract minKey (Succ kChild kChildren) ts)
   = Extract minKey kChildren (Cons kChild ts)
 
 incrExtract' :: LEq a -> BinomTree rk a -> Extract (Succ rk) a -> Extract rk a
-incrExtract' (<=) t (Extract minKey (Succ kChild kChildren) ts)
-  = Extract minKey kChildren (Skip (incr (<=) (t `cat` kChild) ts))
+incrExtract' le t (Extract minKey (Succ kChild kChildren) ts)
+  = Extract minKey kChildren (Skip (incr le (t `cat` kChild) ts))
   where
-    cat = joinBin (<=)
+    cat = joinBin le
 
 -- | Walks backward from the biggest key in the forest, as far as rank @rk@.
 -- Returns its progress.  Each successive application of @extractBin@ takes
 -- amortized /O(1)/ time, so applying it from the beginning takes /O(log n)/ time.
 extractBin :: LEq a -> BinomForest rk a -> MExtract rk a
 extractBin _ Nil = No
-extractBin (<=) (Skip f) = case extractBin (<=) f of
+extractBin le (Skip f) = case extractBin le f of
   Yes ex -> Yes (incrExtract ex)
   No     -> No
-extractBin (<=) (Cons t@(BinomTree x ts) f) = Yes $ case extractBin (<=) f of
+extractBin le (Cons t@(BinomTree x ts) f) = Yes $ case extractBin le f of
   Yes ex@(Extract minKey _ _)
-    | minKey < x  -> incrExtract' (<=) t ex
-  _               -> Extract x ts (Skip f)
-  where a < b = not (b <= a)
+    | minKey `lt` x -> incrExtract' le t ex
+  _                 -> Extract x ts (Skip f)
+  where a `lt` b = not (b `le` a)
 
 mapMaybeQueue :: (a -> Maybe b) -> LEq b -> (rk a -> MinQueue b) -> MinQueue b -> BinomForest rk a -> MinQueue b
-mapMaybeQueue f (<=) fCh q0 forest = q0 `seq` case forest of
+mapMaybeQueue f le fCh q0 forest = q0 `seq` case forest of
   Nil    -> q0
-  Skip forest'  -> mapMaybeQueue f (<=) fCh' q0 forest'
-  Cons t forest'  -> mapMaybeQueue f (<=) fCh' (union' (<=) (mapMaybeT t) q0) forest'
-  where fCh' (Succ t tss) = union' (<=) (mapMaybeT t) (fCh tss)
-        mapMaybeT (BinomTree x ts) = maybe (fCh ts) (\ x -> insert' (<=) x (fCh ts)) (f x)
+  Skip forest'  -> mapMaybeQueue f le fCh' q0 forest'
+  Cons t forest'  -> mapMaybeQueue f le fCh' (union' le (mapMaybeT t) q0) forest'
+  where fCh' (Succ t tss) = union' le (mapMaybeT t) (fCh tss)
+        mapMaybeT (BinomTree x0 ts) = maybe (fCh ts) (\ x -> insert' le x (fCh ts)) (f x0)
 
 type Partition a b = (MinQueue a, MinQueue b)
 
 mapEitherQueue :: (a -> Either b c) -> LEq b -> LEq c -> (rk a -> Partition b c) -> Partition b c ->
   BinomForest rk a -> Partition b c
-mapEitherQueue f (<=) (<=.) fCh (q0, q1) ts = q0 `seq` q1 `seq` case ts of
-  Nil        -> (q0, q1)
-  Skip ts'   -> mapEitherQueue f (<=) (<=.) fCh' (q0, q1) ts'
-  Cons t ts' -> mapEitherQueue f (<=) (<=.) fCh' (both (union' (<=)) (union' (<=.)) (partitionT t) (q0, q1)) ts'
+mapEitherQueue f0 leB leC fCh (q00, q10) ts0 = q00 `seq` q10 `seq` case ts0 of
+  Nil        -> (q00, q10)
+  Skip ts'   -> mapEitherQueue f0 leB leC fCh' (q00, q10) ts'
+  Cons t ts' -> mapEitherQueue f0 leB leC fCh' (both (union' leB) (union' leC) (partitionT t) (q00, q10)) ts'
   where  both f g (x1, x2) (y1, y2) = (f x1 y1, g x2 y2)
-         fCh' (Succ t tss) = both (union' (<=)) (union' (<=.)) (partitionT t) (fCh tss)
+         fCh' (Succ t tss) = both (union' leB) (union' leC) (partitionT t) (fCh tss)
          partitionT (BinomTree x ts) = case fCh ts of
-           (q0, q1) -> case f x of
-             Left b  -> (insert' (<=) b q0, q1)
-             Right c  -> (q0, insert' (<=.) c q1)
+           (q0, q1) -> case f0 x of
+             Left b  -> (insert' leB b q0, q1)
+             Right c  -> (q0, insert' leC c q1)
 
 {-# INLINE tip #-}
 -- | Constructs a binomial tree of rank 0.
@@ -352,46 +361,46 @@ insertMin (BinomTree x ts) (Cons t' f) = Skip (insertMin (BinomTree x (Succ t' t
 -- Each successive application of this function costs /O(1)/, so applying it
 -- from the beginning costs /O(log n)/.
 merge :: LEq a -> BinomForest rk a -> BinomForest rk a -> BinomForest rk a
-merge (<=) f1 f2 = case (f1, f2) of
-  (Skip f1', Skip f2')    -> Skip (merge (<=) f1' f2')
-  (Skip f1', Cons t2 f2') -> Cons t2 (merge (<=) f1' f2')
-  (Cons t1 f1', Skip f2') -> Cons t1 (merge (<=) f1' f2')
+merge le f1 f2 = case (f1, f2) of
+  (Skip f1', Skip f2')    -> Skip (merge le f1' f2')
+  (Skip f1', Cons t2 f2') -> Cons t2 (merge le f1' f2')
+  (Cons t1 f1', Skip f2') -> Cons t1 (merge le f1' f2')
   (Cons t1 f1', Cons t2 f2')
-        -> Skip (carry (<=) (t1 `cat` t2) f1' f2')
+        -> Skip (carry le (t1 `cat` t2) f1' f2')
   (Nil, _)                -> f2
   (_, Nil)                -> f1
-  where  cat = joinBin (<=)
+  where  cat = joinBin le
 
--- | Merges two binomial forests with another tree. If we are thinking of the trees 
+-- | Merges two binomial forests with another tree. If we are thinking of the trees
 -- in the binomial forest as binary digits, this corresponds to a carry operation.
 -- Each call to this function takes /O(1)/ time, so in total, it costs /O(log n)/.
 carry :: LEq a -> BinomTree rk a -> BinomForest rk a -> BinomForest rk a -> BinomForest rk a
-carry (<=) t0 f1 f2 = t0 `seq` case (f1, f2) of
-  (Skip f1', Skip f2')    -> Cons t0 (merge (<=) f1' f2')
+carry le t0 f1 f2 = t0 `seq` case (f1, f2) of
+  (Skip f1', Skip f2')    -> Cons t0 (merge le f1' f2')
   (Skip f1', Cons t2 f2') -> Skip (mergeCarry t0 t2 f1' f2')
   (Cons t1 f1', Skip f2') -> Skip (mergeCarry t0 t1 f1' f2')
   (Cons t1 f1', Cons t2 f2')
         -> Cons t0 (mergeCarry t1 t2 f1' f2')
-  (Nil, _f2)              -> incr (<=) t0 f2
-  (_f1, Nil)              -> incr (<=) t0 f1
-  where  cat = joinBin (<=)
-         mergeCarry tA tB = carry (<=) (tA `cat` tB)
+  (Nil, _f2)              -> incr le t0 f2
+  (_f1, Nil)              -> incr le t0 f1
+  where  cat = joinBin le
+         mergeCarry tA tB = carry le (tA `cat` tB)
 
 -- | Merges a binomial tree into a binomial forest.  If we are thinking
 -- of the trees in the binomial forest as binary digits, this corresponds
 -- to adding a power of 2.  This costs amortized /O(1)/ time.
 incr :: LEq a -> BinomTree rk a -> BinomForest rk a -> BinomForest rk a
-incr (<=) t f = t `seq` case f of
+incr le t f0 = t `seq` case f0 of
   Nil  -> Cons t Nil
   Skip f     -> Cons t f
-  Cons t' f' -> Skip (incr (<=) (t `cat` t') f')
-  where  cat = joinBin (<=)
+  Cons t' f' -> Skip (incr le (t `cat` t') f')
+  where  cat = joinBin le
 
 -- | The carrying operation: takes two binomial heaps of the same rank @k@
 -- and returns one of rank @k+1@.  Takes /O(1)/ time.
 joinBin :: LEq a -> BinomTree rk a -> BinomTree rk a -> BinomTree (Succ rk) a
-joinBin (<=) t1@(BinomTree x1 ts1) t2@(BinomTree x2 ts2)
-  | x1 <= x2  = BinomTree x1 (Succ t2 ts1)
+joinBin le t1@(BinomTree x1 ts1) t2@(BinomTree x2 ts2)
+  | x1 `le` x2 = BinomTree x1 (Succ t2 ts1)
   | otherwise  = BinomTree x2 (Succ t1 ts2)
 
 instance Functor Zero where
@@ -430,13 +439,13 @@ instance Foldable rk => Foldable (BinomForest rk) where
 
 -- instance Traversable Zero where
 --   traverse _ _ = pure Zero
--- 
+--
 -- instance Traversable rk => Traversable (Succ rk) where
 --   traverse f (Succ t ts) = Succ <$> traverse f t <*> traverse f ts
--- 
+--
 -- instance Traversable rk => Traversable (BinomTree rk) where
 --   traverse f (BinomTree x ts) = BinomTree <$> f x <*> traverse f ts
--- 
+--
 -- instance Traversable rk => Traversable (BinomForest rk) where
 --   traverse _ Nil = pure Nil
 --   traverse f (Skip tss) = Skip <$> traverse f tss
@@ -476,7 +485,7 @@ keysQueue Prio.Empty = Empty
 keysQueue (Prio.MinPQ n k _ ts) = MinQueue n k (keysF (const Zero) ts)
 
 keysF :: (pRk k a -> rk k) -> Prio.BinomForest pRk k a -> BinomForest rk k
-keysF f ts = case ts of
+keysF f ts0 = case ts0 of
   Prio.Nil       -> Nil
   Prio.Skip ts'  -> Skip (keysF f' ts')
   Prio.Cons (Prio.BinomTree k _ ts) ts'
