@@ -21,7 +21,7 @@
 -- use 'seqSpine'.
 --
 -- This implementation does not guarantee stable behavior.
--- 
+--
 -- This implementation offers a number of methods of the form @xxxU@, where @U@ stands for
 -- unordered.  No guarantees whatsoever are made on the execution or traversal order of
 -- these functions.
@@ -31,7 +31,7 @@ module Data.PQueue.Min (
   -- * Basic operations
   empty,
   null,
-  size, 
+  size,
   -- * Query operations
   findMin,
   getMin,
@@ -41,6 +41,7 @@ module Data.PQueue.Min (
   -- * Construction operations
   singleton,
   insert,
+  insertBehind,
   union,
   unions,
   -- * Subsets
@@ -84,13 +85,9 @@ module Data.PQueue.Min (
 
 import Prelude hiding (null, foldr, foldl, take, drop, takeWhile, dropWhile, splitAt, span, break, (!!), filter, map)
 
-import Control.Applicative (Applicative(..), (<$>))
-import Control.Applicative.Identity
-
-import Data.Monoid
-import Data.Maybe hiding (mapMaybe)
+import Data.Monoid (Monoid(mempty, mappend, mconcat))
 import Data.Foldable (foldl, foldr, foldl')
-import Data.Traversable
+import Data.Maybe (fromMaybe)
 
 import qualified Data.List as List
 
@@ -100,13 +97,12 @@ import Data.PQueue.Internals
 import GHC.Exts (build)
 import Text.Read (Lexeme(Ident), lexP, parens, prec,
   readPrec, readListPrec, readListPrecDefault)
-import Data.Data
 #else
 build :: ((a -> [a] -> [a]) -> [a] -> [a]) -> [a]
 build f = f (:) []
 #endif
 
--- instance 
+-- instance
 
 instance (Ord a, Show a) => Show (MinQueue a) where
   showsPrec p xs = showParen (p > 10) $
@@ -150,7 +146,7 @@ deleteFindMin = fromMaybe (error "Error: deleteFindMin called on empty queue") .
 unions :: Ord a => [MinQueue a] -> MinQueue a
 unions = foldl union empty
 
--- | /O(k log n)/.  Index (subscript) operator, starting from 0.  @queue !! k@ returns the @(k+1)@th smallest 
+-- | /O(k log n)/.  Index (subscript) operator, starting from 0.  @queue !! k@ returns the @(k+1)@th smallest
 -- element in the queue.  Equivalent to @toAscList queue !! k@.
 (!!) :: Ord a => MinQueue a -> Int -> a
 q !! n  | n >= size q
@@ -166,11 +162,11 @@ takeWhile p = foldWhileFB p . toAscList
 {-# INLINE foldWhileFB #-}
 -- | Equivalent to Data.List.takeWhile, but is a better producer.
 foldWhileFB :: (a -> Bool) -> [a] -> [a]
-foldWhileFB p xs = build (\ c nil -> let 
+foldWhileFB p xs0 = build (\ c nil -> let
   consWhile x xs
     | p x    = x `c` xs
     | otherwise  = nil
-  in foldr consWhile nil xs)
+  in foldr consWhile nil xs0)
 
 -- | 'dropWhile' @p queue@ returns the queue remaining after 'takeWhile' @p queue@.
 dropWhile :: Ord a => (a -> Bool) -> MinQueue a -> MinQueue a
@@ -184,7 +180,7 @@ dropWhile p = drop' where
 -- satisfy @p@ and second element is the remainder of the queue.
 span :: Ord a => (a -> Bool) -> MinQueue a -> ([a], MinQueue a)
 span p queue = case minView queue of
-  Just (x, q') 
+  Just (x, q')
     | p x  -> let (ys, q'') = span p q' in (x:ys, q'')
   _        -> ([], queue)
 
@@ -240,8 +236,8 @@ toDescList :: Ord a => MinQueue a -> [a]
 toDescList queue = build (\ c nil -> foldrDesc c nil queue)
 
 {-# INLINE toList #-}
--- | /O(n)/.  Returns the elements of the priority queue in ascending order.  Equivalent to 'toAscList'.
--- 
+-- | /O(n log n)/.  Returns the elements of the priority queue in ascending order.  Equivalent to 'toAscList'.
+--
 -- If the order of the elements is irrelevant, consider using 'toListU'.
 toList :: Ord a => MinQueue a -> [a]
 toList = toAscList
@@ -291,7 +287,7 @@ mapU = mapMonotonic
 elemsU :: MinQueue a -> [a]
 elemsU = toListU
 
--- | Returns the elements of the queue, in no particular order.
+-- | /O(n)/.  Returns the elements of the queue, in no particular order.
 toListU :: MinQueue a -> [a]
 toListU q = build (\ c n -> foldrU c n q)
 
