@@ -20,13 +20,13 @@
 -- This implementation is based on a binomial heap augmented with a global root.
 -- The spine of the heap is maintained lazily.  To force the spine of the heap,
 -- use 'seqSpine'.
--- 
+--
 -- We do not guarantee stable behavior.
--- Ties are broken arbitrarily -- that is, if @k1 <= k2@ and @k2 <= k1@, then there 
+-- Ties are broken arbitrarily -- that is, if @k1 <= k2@ and @k2 <= k1@, then there
 -- are no guarantees about the relative order in which @k1@, @k2@, and their associated
 -- elements are returned.  (Unlike Data.Map, we allow multiple elements with the
 -- same key.)
--- 
+--
 -- This implementation offers a number of methods of the form @xxxU@, where @U@ stands for
 -- unordered.  No guarantees whatsoever are made on the execution or traversal order of
 -- these functions.
@@ -37,8 +37,9 @@ module Data.PQueue.Prio.Min (
   empty,
   singleton,
   insert,
+  insertBehind,
   union,
-  unions, 
+  unions,
   -- * Query
   null,
   size,
@@ -115,22 +116,23 @@ module Data.PQueue.Prio.Min (
   )
   where
 
-import Control.Applicative (Applicative (..), (<$>))
-import Data.Monoid 
+import Control.Applicative (Applicative, pure, (<*>), (<$>))
+
 import qualified Data.List as List
-import Data.Foldable (Foldable, foldl, foldr, foldl')
-import Data.Traversable
+import qualified Data.Foldable as Fold(Foldable(..))
+import Data.Monoid (Monoid(mempty, mappend, mconcat))
+import Data.Traversable (Traversable(traverse))
+import Data.Foldable (Foldable)
 import Data.Maybe (fromMaybe)
 
 import Data.PQueue.Prio.Internals
 
-import Prelude hiding (map, filter, break, span, takeWhile, dropWhile, splitAt, take, drop, (!!), null, foldr)
+import Prelude hiding (map, filter, break, span, takeWhile, dropWhile, splitAt, take, drop, (!!), null)
 
 #ifdef __GLASGOW_HASKELL__
 import GHC.Exts (build)
 import Text.Read (Lexeme(Ident), lexP, parens, prec,
   readPrec, readListPrec, readListPrecDefault)
-import Data.Data
 #else
 build :: ((a -> [a] -> [a]) -> [a] -> [a]) -> [a]
 build f = f (:) []
@@ -211,7 +213,7 @@ mapKeys f q = fromList [(f k, a) | (k, a) <- toListU q]
 
 -- | /O(n log n)/.  Traverses the elements of the queue in ascending order by key.
 -- (@'traverseWithKey' f q == 'fromAscList' <$> 'traverse' ('uncurry' f) ('toAscList' q)@)
--- 
+--
 -- If you do not care about the /order/ of the traversal, consider using 'traverseWithKeyU'.
 traverseWithKey :: (Ord k, Applicative f) => (k -> a -> f b) -> MinPQueue k a -> f (MinPQueue k b)
 traverseWithKey f q = case minViewWithKey q of
@@ -252,10 +254,10 @@ take n = List.take n . toAscList
 
 -- | /O(k log n)/.  Deletes the first @k@ (key, value) pairs in the queue, or returns an empty queue if @k >= n@.
 drop :: Ord k => Int -> MinPQueue k a -> MinPQueue k a
-drop n q 
-  | n <= 0  = q
-  | n >= size q  = empty
-  | otherwise  = drop' n q
+drop n0 q0
+  | n0 <= 0  = q0
+  | n0 >= size q0  = empty
+  | otherwise  = drop' n0 q0
   where
     drop' n q
       | n == 0    = q
@@ -263,7 +265,7 @@ drop n q
 
 -- | /O(k log n)/.  Equivalent to @('take' k q, 'drop' k q)@.
 splitAt :: Ord k => Int -> MinPQueue k a -> ([(k, a)], MinPQueue k a)
-splitAt n q 
+splitAt n q
   | n <= 0     = ([], q)
   | otherwise  = n `seq` case minViewWithKey q of
       Just (ka, q') -> let (kas, q'') = splitAt (n-1) q' in (ka:kas, q'')
@@ -279,7 +281,7 @@ takeWhile = takeWhileWithKey . const
 -- | Takes the longest possible prefix of elements satisfying the predicate.
 -- (@'takeWhile' p q == 'List.takeWhile' (uncurry p) ('toAscList' q)@)
 takeWhileWithKey :: Ord k => (k -> a -> Bool) -> MinPQueue k a -> [(k, a)]
-takeWhileWithKey p = takeWhileFB (uncurry' p) . toAscList where
+takeWhileWithKey p0 = takeWhileFB (uncurry' p0) . toAscList where
   takeWhileFB p xs = build (\ c n -> foldr (\ x z -> if p x then x `c` z else n) n xs)
 
 -- | Removes the longest possible prefix of elements satisfying the predicate.
@@ -320,10 +322,10 @@ fromAscList = foldr (uncurry' insertMin) empty
 
 -- | /O(n)/.  Build a priority queue from a descending list of (key, value) pairs.  /The precondition is not checked./
 fromDescList :: [(k, a)] -> MinPQueue k a
-fromDescList = foldl' (\ q (k, a) -> insertMin k a q) empty
+fromDescList = List.foldl' (\ q (k, a) -> insertMin k a q) empty
 
 {-# RULES
-  "fromList/build" forall (g :: forall b . ((k, a) -> b -> b) -> b -> b) . 
+  "fromList/build" forall (g :: forall b . ((k, a) -> b -> b) -> b -> b) .
     fromList (build g) = g (uncurry' insert) empty;
   "fromAscList/build" forall (g :: forall b . ((k, a) -> b -> b) -> b -> b) .
     fromAscList (build g) = g (uncurry' insertMin) empty;
@@ -355,7 +357,7 @@ toDescList = foldlWithKey (\ z k a -> (k, a) : z) []
 
 {-# INLINE toList #-}
 -- | /O(n log n)/.  Equivalent to 'toAscList'.
--- 
+--
 -- If the traversal order is irrelevant, consider using 'toListU'.
 toList :: Ord k => MinPQueue k a -> [(k, a)]
 toList = toAscList
